#!/local/paths/perl -wT
#
# File: authstats.pl
#
# Description: reads authentication log file and extracts connection
#              statistics.
#
# Author: Sylvain Robitaille
#
# Date: 1998/10/14: begin
#       1998/11/02: Fix line parsing bug. (Dates with day < 10 were being
#                   incorrectly split)
#       1998/12/01: We now can recognize 'shutdown' lines in the log.
#                   We now do a lookup on hostnames if the mechanism was
#                   ssh.
#       1998/12/04: Add options to sort lists by number of connections
#                   or by host name.
#       1999/06/16: Support logs from multiple hosts. Default to
#                   reporting only on entries from the local host (as
#                   reported by hostname), with command line option to
#                   report on a selected host's entries.
#       1999/09/30: Ssh-1.2.27 now logs connections the same way as
#                   other services, so we no longer need special code
#                   to process that. We still ignore other entries
#                   sshd, though.
#                   We now recognize 'twist' as an accepted connection.
#       2000/05/15: 'twist' is neither accepted nor denied...
#                   We now provide an "executive" summary of all
#                   connections, by service before the full report.
#                   The report generation has been simplified.
#                   Comments have been cleaned up.
#       2000/05/16: Combine "ignore list" entries from all our hosts.
#       2001/03/26: Handle uninformative shutdown log from Solaris.
#                   Separator line between the executive summary and the
#                   full report.
#

#
# First some directives for the compiler.
use strict;            # Everybody should use this, IM(NS)HO
use Getopt::Std;       # command line arguments
use File::Basename;    # filename parsing
use Sys::Syslog;       # syslog logging
use Socket;            # network sockets


####################################################################
#
# CONFIGURATION section. Edit to your heart's content!

# The syslog facility to use if running as uid = 0:
my $root_syslog_facility = 'syslog';

# The syslog facility to use if running as any other user. This may or
# may not be the same as $root_syslog_facility.
my $user_syslog_facility = 'user';

# The program to use for mailing errors. Note that this program must
# recognize a command line of the form 'program -s subject user'. The
# user is determined by our -m command-line option. Without -m, errors
# will not be mailed.
my $mail_program = 'mailx';

# The subject we'll use when mailing our report.
my $subject = "Connection stats";

# Environment variables. Keep these to known safe values.
$ENV{PATH} = '/local/paths:/usr/bin:/bin';
$ENV{IFS}  = " \t\n";    # <space><tab><newline>
umask 077;


# The default input file to read from
my $input_file = "/var/adm/syslog/authlog";

# The default host whose logs we're interested in
my $target_host = `hostname`;

#
# End of configuration.
####################################################################

#
# "Option" variables.
use vars qw($opt_c $opt_d $opt_f $opt_h $opt_m $opt_q $opt_t);

# They translate to:
my $connsort  = 0;
my $debug     = 0;
my $hostsort  = 0;
my $mail_rcpt = 0;
my $quiet     = 0;

# If set, $opt_f overrides $input_file  default.
# If set, $opt_t overrides $target_host default.

#
# Format for printed report

my $hostname;
my $connections;

format REPORT =
    @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<    @>>>>>
    $hostname, $connections
.


# SUBROUTINE: fatal_error( $message )
#
# Description: executes a call to die, with $message as an
#              argument.
#
# NOTE: We want to syslog the error message before dying.
#
# Author: Sylvain Robitaille
#
# Date: June 19, 1997
#
#       1998/02/04: include logging via syslog.
#
sub fatal_error {
    syslog 'err', "FATAL error: @_\n";
    die "FATAL error: @_\n";
}

# SUBROUTINE: debug_message( $message )
#
# Description: Prints $message to STDOUT, clearly
#              identified as debug information. Prints only 
#              if $debug is a true value.
#
# Author: Sylvain Robitaille
#
# Date: July 22, 1997
#
#       1998/02/04: - include logging via syslog.
#       1998/03/06: - but only if in debug mode!
#
sub debug_message {
  if( $debug ) {
    syslog 'debug', "@_\n";
    print "Debug: @_\n"
  }
}

# SUBROUTINE: error_message( $message )
#
# Description: Prints messages to $errors_file if writable, and to
#              STDERR if not in quiet mode.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#       1998/02/04: - include logging via syslog.
#       1998/02/06: - prefix message with the string 'Error: '
#       1998/09/01: - Include date/time in output.
#
sub error_message {
  my $time = localtime;

  syslog 'err', "Error: @_\n";
  print STDERR  "$time: Error: @_\n" unless $quiet;

# open  ERROR,  ">>$errors_file" or fatal_error "can't open $errors_file!";
# print ERROR   "$time: Error: @_\n";
# close ERROR;
}

# SUBROUTINE: info_message( $message )
#
# Description: Prints messages to STDOUT if not in quiet mode.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/04
#
sub info_message {
  syslog 'info', "@_\n";
  print STDOUT "@_\n" unless $quiet;
}

# SUBROUTINE: untaint_file( $path )
#
# Description: Ensures that no illegal file characters are
#              in $path.
#
# Returns: An untainted copy of $path or the string 'TAINTED'.
#
# NOTES: This currently checks only that characters
#        contained in $path are legal. A more complete
#        untaint mechanism might check that the file
#        pointed to by $path is a real file, and that the
#        user has access to that file, or that the user
#        can create the file if it doesn't exist, etc.
#
# Author: Sylvain Robitaille
#
# Date: June 19, 1997
#
#       1998/02/04: - rather than call it a fatal error if the path is
#                     tainted, return 0. We will check for this and
#                     simply complain and skip a file if it contains a
#                     tainted path.
#       1998/02/06: - return the string 'TAINTED' if the path is
#                     tainted. This will make it easier to check the
#                     return value, by a string comparison rather than
#                     checking for a numeric value.
#       1998/02/17: - '+' is a legal filename character.
#       1999/06/16: - Actually, it's better to return an undefined value
#                     if the string is tainted.
#                   - Adjust the regular expression to any printable
#                     character that isn't a shell meta-character.
#
sub untaint_file {
    my $path = shift @_;

    if($path =~ /^([^\x00-\x20<>`|]+)$/) {
        $path = $1;
    } else {
        undef $path;
    }
    return $path;
}

# SUBROUTINE: untaint_mail( $path )
#
# Description: Ensures that no illegal file characters are
#              in $path.
#
# Returns: An untainted copy of $path or the string 'TAINTED'.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/19: - copy from untaint_file and adjust.
#       1998/02/20: - allow whitespace to permit multiple recipients.
#       1999/06/16: - return an undefined value if the string is tainted.
#
sub untaint_mail {
    my $path = shift @_;

    if($path =~ /^([\s\-@\w\.~\+]+)$/) {
        $path = $1;
    } else {
        undef $path;
    }
    return $path;
}

# Subroutine: Usage
#
# Description: Prints a usage message.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#       1998/02/04: - Add '-m recipient' option.
#       1998/12/04: - Add sorting options. Describe options.
#
sub Usage {
  my $progname = basename($0, ());

  print STDERR <<USAGE;
    Usage: $progname [-d][-q][-c|-h][-f <file>][-m <recipient>][-t <host>]

    -d: Print debug messages.
    -q: Do not print "informational" status messages.
    -c: Sort lists by the number of connections.
    -h: Sort lists by host name/address.
    -f: Use file for input.
    -m: Email output to recipient.
    -t: Report on entries from host.
USAGE

exit 1;
}

# Subroutine: Parse_Options
#
# Description: Parses the command line options into default global
#              variables.
#
# Uses globals: opt_c
#               opt_d
#               opt_f
#               opt_h
#               opt_m
#               opt_q
#
# Author: Sylvain Robitaille
#
# Date:
#
#       1998/02/04: - Add '-m recipient' option to define who gets the
#                     mailed errors file.
#       1998/02/06: - Changed check of return value from untaint_file to
#                     a string comparison, rather than checking for a
#                     numeric return value.
#       1998/02/19: - The mail recipient must also be untainted.
#       1998/12/04: - Sort options added: 'c' sort by number of
#                     connections, and 'h' sort by hostname.
#       1999/06/16: - Add option to select which host's entries we're
#                     after.
#
sub Parse_Options {

  if(! getopts('cdf:hm:qt:')) {
    Usage;
  }

  $connsort    = $opt_c if( $opt_c );
  $debug       = $opt_d if( $opt_d );
  $quiet       = $opt_q if( $opt_q );
  $hostsort    = $opt_h if( $opt_h );

  $mail_rcpt   = untaint_mail $opt_m if( $opt_m );
  if( ! defined $mail_rcpt ) {
    fatal_error "Tainted email address: \'$opt_m\'";
  }
  $input_file = untaint_file $opt_f if( $opt_f );
  if( ! defined $input_file ) {
    fatal_error "Tainted path to file: \'$opt_f\'";
  }
  $target_host = untaint_file $opt_t if( $opt_t );
  if( ! defined $target_host ) {
    fatal_error "Tainted hostname: \'$opt_t\'";
  }
  chomp $target_host;
  debug_message "Target host is:\'$target_host\'";

  # Sanity check: We can only sort by connections, or by host, but not
  # by both.
  if( $connsort && $hostsort ) {
    Usage;
  }
}

sub FindHost {
    my $remote = shift;
    my ($remote_user, $remote_host);

    ($remote_user, $remote_host) = split /@/, $remote;
    $remote_host = $remote_user if( !defined $remote_host );
    return $remote_host;
}

# Subroutine: dns_search_addr $ip
#
# Description: Performs a DNS search on the IP address $ip.
#
# Author: Sylvain Robitaille
#
# Date: 1998/12/01
#
sub dns_search_addr {
    my $ip = shift;
    my $addr;

    $addr = pack 'C4', $ip =~ m%(\d+)\.(\d+)\.(\d+)\.(\d+)%;
    $addr = gethostbyaddr( $addr, AF_INET );

    return $addr;
}

# Subroutine: sortkeys \%cncts
# 
# Description: sorts elements in the %cncts hash, according to method
#              selected on the command line. (Defaults to no sorting)
# 
# Author: Sylvain Robitaille
# 
# Date: 1998/12/04
#
sub sortkeys {
    my $cncts = shift;

    if( $hostsort ) {
        return sort keys %$cncts;
    } elsif( $connsort ) {
        return reverse sort { $$cncts{$a} <=> $$cncts{$b} } keys %$cncts;
    } else {
        return keys %$cncts;
    }
}

# The main routine.
#
# Author: Sylvain Robitaille
#
# Date: 1998/10/14
#       1998/12/01: We now can recognize 'shutdown' lines in the log.
#

# Variables used by main.
my $facility;        # Which facility will we log to syslog with?
my $progname;        # Who are we, anyway?
my $oldformat;
my $oldfh;

my ($service, $action, $remote, $remote_user, $remote_host, $host_ip);
my $rest;
my @parts;
my $log_host;

# cumulative connection counts:
my %cumulative;

# connection details:
my %conn;

# make sure our default $target_host is not tainted
$target_host = untaint_file $target_host;
if( ! defined $target_host ) {
  fatal_error "Unable to determine local host name. Got `hostname`";
}

# Now get command line options.
Parse_Options;

# Who are we, anyway?
$progname = basename $0;

# Ok we're started up correctly. We want to open up a connection to
# syslog, with the appropriate facility, and log that we're starting.
#
# We have the following options for syslog logging:
#
# pid:    Logs the process ID with each message.
# ndelay: Opens the connect to syslogd immediately, instead of when the
#         first message is logged.
# cons:   Sends messages to the console if unable to send them to
#         syslogd.
# nowait: Logs messages to the console without waiting for child
#         processes that are forked.
#
# $> is our effective user id.
if( $> == 0 ) {
  $facility = $root_syslog_facility;
} else {
  $facility = $user_syslog_facility;
}
openlog $progname, 'pid', $facility;

# Ok, here we go...
info_message "$progname started.";

# Open the data file for reading
open DATAFILE, "<$input_file" or
    fatal_error "Can't open input file: $input_file: $!";
while( <DATAFILE> ) {
    chomp;
    # Parse the line, and add each item to the appropriate hash
    (undef, undef, undef, $log_host, $service, $action, $rest) = split m% +%, $_, 7;

    # We're only interested in lines that match $target_host
    next if( $log_host !~ m%^($target_host)$%i );

    $service =~ s/(.*)\[\d+\]:/$1/;
    #
    # Ignore the following "services"
    next if( $service eq 'newshell'    || $service eq 'identd'     ||
             $service eq 'sendmail'    || $service eq 'field'      ||
             $service eq 'ntalkd'      || $service eq 'talkd'      ||
             $service eq 'pw_wrapper'  || $service eq 'yapasswd'   ||
             $service eq 'comsat'      || $service eq 'last'       ||
             $service eq 'su:'         || $service eq 'xlock'      ||
             $service eq 'qi.start'    || $service eq 'enabler'    ||
             $service eq 'idle_finder' || $service eq 'monitor-su' ||
             $service eq 'login:'      || $service eq 'mailhandler.pl' );
    #
    # Need a case for $action eq 'warning' in which case the line is a
    # message from tcp_wrappers stating that a connection was dropped
    # because the DNS lookup failed.
    #
    if( $action eq 'connect' ) {
        (undef, $remote) = split /\s+/, $rest;
        $remote_host = FindHost( $remote );
        $conn{'ACCEPT'}{$service}{$remote_host}++;
        #
        # keep a running total of connections:
        $cumulative{'ACCEPT'}{$service}++;

    } elsif( $action eq 'refused' ) {
        #
        # Fix me: FTP needs to be special cased, because it logs so
        # much, such as refused commands.
        #
        (undef, undef, $remote) = split /\s+/, $rest;
        $remote_host = FindHost( $remote );
        $conn{'REFUS'}{$service}{$remote_host}++;
        #
        # keep a running total of connections:
        $cumulative{'REFUS'}{$service}++;

    } elsif( $action eq 'twist' ) {
        ($remote) = split /\s+/, $rest;
        $remote_host = FindHost( $remote );
        $conn{'TWIST'}{$service}{$remote_host}++;
        #
        # keep a running total of connections:
        $cumulative{'TWIST'}{$service}++;

    } elsif( $service eq 'ftpd' ) {
        #
        # ignore ftpd log entries that aren't 'connect' or 'refused'
        next;

    } elsif( $service eq 'sshd' ) {
         #
         # New since ssh-1.2.27: Other lines are not going to give us
         # any particularly interesting information, since we have the
         # hostname already.
         next;

         # For ssh versions < 1.2.27, we need to look up the IP address
         # if we want to see hostnames in the summary report. Comment
         # out the above 'next;' and uncomment the following lines:
         #
#        @parts = split /\s+/, $rest;
#        if( $parts[0] eq 'Connection' ) {
#
#            # Look up the host name, since sshd only logs the IP address.
#            $host_ip     = $parts[$#parts - 2];
#            $remote_host = dns_search_addr $host_ip;
#
#            # 1999/03/18 Sylvain Robitaille: Allow for the host to
#            #                                not be registered. :-(
#            $remote_host = $host_ip if( ! defined $remote_host );
#
#            $conn{$service}{$remote_host}++;
#        } else {
#            # We don't get any information that is interesting to this
#            # report from other lines.
#            next;
#        }

    } elsif( $action eq 'warning:' ) {
        # not yet implemented warnings about failed DNS lookups.
    } elsif( $service eq 'poppassd' && $action eq 'failed' ) {
        # ignore failed password change attempts.
        next;
    } elsif( $service eq 'shutdown:' ) {
        # 2001/03/26 Sylvain Robitaille: Solaris may not give any
        # additional information about who called shutdown.
        if( defined $action and defined $rest ) {
          $remote_host = "$action $rest";
        } else {
          $remote_host = "unknown (probably Solaris shutdown)";
        }
        $conn{'ACCEPT'}{$service}{$remote_host}++;
    } elsif( $action eq 'ROOT' ) {
        # Not yet implemented Remote shell as root. Note that this will
        # still be logged as an accepted rshd connection, *and* we'll
        # get a record of it from the su monitor.
    } else {
        error_message "unrecognized line: $_";
    }
}
close DATAFILE;

if( $mail_rcpt ) {
    $subject = "\"$target_host: $subject\"";
    open OUTPUT, "| $mail_program -s $subject $mail_rcpt" or
        fatal_error "can't mail to $mail_rcpt: $!";
    $oldfh = select OUTPUT;
}

$oldformat = $~;
$~ = "REPORT";

# 2000/05/15 Sylvain Robitaille: The executive summary:
foreach $action ('REFUS', 'TWIST', 'ACCEPT') {
    if( defined $cumulative{$action} ) {
        print "\n". $action ."ED connections:\n";
        print "    service                                  connections\n";
        print "--------------------------------------------------------\n";
    }
    foreach $hostname (sortkeys \%{$cumulative{$action}}) {
        # Use $hostname to list the service name:
        $connections = $cumulative{$action}{$hostname};
        write;
    }
}

# 2001/03/26 Sylvain Robitaille: Anne requests some sort of separation
#                                between the executive summary and the
#                                full report:
print "\n" . "=" x 68 . "\n";

# The full report:
foreach $action ('REFUS', 'TWIST', 'ACCEPT') {
    foreach $service (keys %{$conn{$action}}) {
        print "\n". $action ."ED $service connections:\n";
        print "    remote host                              connections\n";
        print "--------------------------------------------------------\n";
        foreach $hostname (sortkeys \%{$conn{$action}{$service}}) {
            $connections = $conn{$action}{$service}{$hostname};
            write;
        }
    }
}

$~ = $oldformat;

if( $mail_rcpt ) {
    select $oldfh;
    close OUTPUT;
}

# Log that we're done before closing the connectio to syslog
info_message "$progname done.";
closelog;
